/***

ehchnvol.c -- Error Handler, support for channels and volumes

    Copyright (C) 1997  CSEG, Inc., All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                   

    The GNU General Public License is in the file COPYING which resides
    in the top directory of the distribution tree.

***/

static const char *rcsid =
       "$Id: ehchnvol.c,v 1.1.1.1 1997/09/02 12:18:16 ccobb Exp $";

#include "ehenv.h"

#if defined(HAVE_BE_KERNEL_OS_H)
#include <be/kernel/OS.h>
#include <iostream.h> /* this has to be included early */
#endif

#undef  EHSIMCTRLC
#include "ehchnvol.h"

#include <stdlib.h>
#include <stdio.h>

/***
IPC_IMPLEMENTED is set to 1 if 
some type of ipc (shared memory?)
allows a remote process to control
the behavior of the error handler
in this process.
***/
/* #if ehMSDOS || ehAMIGA || ehBEOS */
#if defined(HAVE_BE_KERNEL_OS_H) || \
   ( defined(HAVE_UNISTD_H) && defined(HAVE_SYS_IPC_H) && \
     defined(HAVE_SYS_SHM_H) )
#define IPC_IMPLEMENTED 1
#else
#define IPC_IMPLEMENTED 0
#endif

const char *apcEhChan[EH_NUMCHAN] =
{
   "unit", "alpha", "beta" /* unit, alpha, beta */
};

const char *apcEhVol[EH_NUMVOL] =
{ 
   "ERR", "WRN", "REM", "DBG" /* errors, warnings, remarks & debug */
};

/*** public functions ***/
#if ehSUNOS
char *shmat  A((int shmid, char *chmaddr, int shmflg));
int   shmctl A((int shmid, int cmd, struct shmid_ds *buf));
int   shmdt  A((char *shmaddr));
int   shmget A((key_t key, int size, int shmflg));
#endif

/* sole access to channel volume level */
short *phEhChnVol = NO_CHNVOL;

#if IPC_IMPLEMENTED
#if defined(HAVE_BE_KERNEL_OS_H)
static area_id AreaId = -1; // the area we will use and delete
#else
static int     iShmId = -1; // the shared memory id
#endif
#endif

/***
ehInitChnVol()
Obtain shared memory for channel volume level.
***/

#if !IPC_IMPLEMENTED
int
ehInitChnVol A1(pid_t, iPID)
{
   static short hChnVol = 0;

   // puts("ehInitChnVol: entering (1)");

   phEhChnVol = &hChnVol;
   NOREF(iPID);

   // puts("ehInitChnVol: leaving (1)");

   return(0);
}

#else /* IPC_IMPLEMENTED */

#if defined(HAVE_BE_KERNEL_OS_H) /* BeOS C++ code */

int           // 0 if successful, else -1
ehInitChnVol( // initialize the shared channel-volume area
   pid_t iPID)  // if 0 create a new area, else find area for given PID
{
   int iErr = 0; // no errors to start with

   // get the current team number
   thread_info ThreadInfo;
   status_t Status = get_thread_info(find_thread(NULL), &ThreadInfo);
   if (Status < B_NO_ERROR)
   {
      cerr << "ehInitChnVol(): Couldn't get thread info." << endl;
      iErr = 1; goto Exit;
   }

   char acAreaNameOrg[B_OS_NAME_LENGTH];
   if (iPID == 0)/* use current PID to create new area */
   {
      sprintf(acAreaNameOrg, "eh-%u", ThreadInfo.team);
      AreaId = create_area(
            acAreaNameOrg,               // name you give to the area
            (void **)&phEhChnVol,        // returns the starting addr
            B_ANY_ADDRESS,               // area can start anywhere
            B_PAGE_SIZE,                 // size in bytes
            B_NO_LOCK,                   // Lock in RAM?  No.
            B_READ_AREA | B_WRITE_AREA); // permissions
      if (AreaId < B_NO_ERROR)
      {
         cerr << "ehInitChnVol(): Couldn't create new area." << endl;
         iErr = 1; goto Exit;
      }
      *phEhChnVol = EH_VOL_ERR;
   }
   else  /* use given PID to find area for that PID */
   {
      sprintf(acAreaNameOrg, "eh-%u", iPID);
      area_id AreaIdOrg = find_area(acAreaNameOrg); // original area
      if (AreaIdOrg < B_NO_ERROR)
      {
         cerr << "ehInitChnVol(): Couldn't find original area." << endl;
         iErr = 1; goto Exit;
      }
      char acAreaNameClone[B_OS_NAME_LENGTH];
      sprintf(acAreaNameClone, "eh-%u", ThreadInfo.team);
      AreaId = clone_area(              // AreaId is file static
            acAreaNameClone,            // name for new cloned area
            (void**)&phEhChnVol,        // where to put area address
            B_ANY_ADDRESS,              // can start anywhere
            B_READ_AREA | B_WRITE_AREA, // read/write access
            AreaIdOrg);                 // what area to clone
      if (AreaId < B_NO_ERROR)
      {
         cerr << "ehInitChnVol(): Couldn't clone area." << endl;
         iErr = 1; goto Exit;
      }
   }

Exit:
   if (iErr)
   {
      ehTermChnVol();
   }
   return(iErr);
}

#else /* *nix probably */

int
ehInitChnVol A1(pid_t, iPID)
/* {PID|0} if 0, use current process id and IPC_CREAT */
/*         if PID, use it and DON'T use IPC_CREAT */
{
   int iErr = 0;
   key_t ktKey;
   int iMode;

#define SZ_SHMEM MAX(sizeof(*phEhChnVol),MINSHMEM)

   // puts("ehInitChnVol: entering (2)");

   if (iPID == EH_NEWDBGVAL)/* use current PID and create new dbgval */
   {
      iPID = getpid();
      iMode = IPC_CREAT | 0660;
   }
   else  /* use given PID, find dbgval for that PID (no IPC_CREAT) */
   {
      iMode = 0660;
   }

   ktKey = (key_t)(EH_PIDPREFIX + iPID);
   //fprintf(stderr, "ehInitChnVol: calling shmget(%x, %d, %d)\n", 
   //      ktKey, SZ_SHMEM, iMode);
   if ((iShmId = shmget(ktKey, SZ_SHMEM, iMode)) == -1)
   {
      perror("ehchnvol.c->shmget()");
      fprintf(stderr, "shmget() failed for pid %d\n", iPID);
      fflush(stderr);
      iErr = 1;
      goto Exit;
   }

   if ((phEhChnVol = (short *)shmat(iShmId, (char *)NULL, 0))
         == NO_CHNVOL)
   {
      perror("ehInitChnVol->shmat");
      iErr = 1;
      goto Exit;
   }

   if (iMode & IPC_CREAT)
   {
      *phEhChnVol = EH_VOL_ERR;
   }

Exit:
   if (iErr)
   {
      ehTermChnVol();
   }

   // puts("ehInitChnVol: leaving (2)");

   return(iErr);
}

#endif /* defined(HAVE_BE_KERNEL_OS_H) */
#endif /* IPC_IMPEMENTED */

/***
ehTermChnVol()
Obtain shared memory for channel volume level.
***/

#if !IPC_IMPLEMENTED

void ehTermChnVol A0()
{
}

#else /* IPC_IMPLEMENTED */

#if defined(HAVE_BE_KERNEL_OS_H) /* BeOS C++ code */

void
ehTermChnVol() // detach from or delete area
{
   if (phEhChnVol != NO_CHNVOL && AreaId > 0)
   {
      delete_area(AreaId);
      phEhChnVol = NO_CHNVOL;
   }
   AreaId = -1;
Exit:
   return;
}

#else /* *nix probably */

void
ehTermChnVol()
{
   if (phEhChnVol != NO_CHNVOL)
   {
      if (shmdt((char *)phEhChnVol))
      {
         fprintf(stderr, "shmdt() failed\n");
         fflush(stderr);
         goto Exit;
      }
      phEhChnVol = NO_CHNVOL;
   }
   if (iShmId != -1) // we obtained a shared memory id
   {
      struct shmid_ds ShmIdDs; // Shared mem data structure
      if (shmctl(iShmId, IPC_STAT, &ShmIdDs))
      {
         perror("smctl(IPC_STAT) failed");
         goto Exit;
      }
      if (ShmIdDs.shm_cpid == getpid()) // if we created the shared mem
      {
         if (shmctl(iShmId, IPC_RMID, NULL)) // then delete it
         {
            perror("smctl(IPC_RMID) failed");
            goto Exit;
         }
      }
      iShmId = -1;
   }
Exit:
   return;
}

#endif /* defined(HAVE_BE_KERNEL_OS_H) */
#endif /* IPC_IMPLEMENTED */


#ifndef ehGetVol

/***
ehGetVol()
Get channel volume level from shared memory.
***/

int 
ehGetVol A1(int, iChan)   /* {0..EH_NUMCHAN} which channel to change */
{
   int iVol = EH_VOL_ERR; /* default to ERR volume */

   if (phEhChnVol == NO_CHNVOL)
   {
      goto Exit;
   }

   /*** get the channel priority ***/
   iVol = (int)(*phEhChnVol >> (EH_BITS_PER_CHAN * iChan)) & EH_VOL_MASK;

Exit:
   return(iVol);
}

#endif

/***
ehSetVol()
Set channel volume level in shared memory.
***/

int 
ehSetVol A2(int, iChan, 
            int, iVol)
             /* {0..15} which channel to change */
             /* {0..3,-1} new volumes for channel */
{
   int iErr = 0;

   if (phEhChnVol == NO_CHNVOL)
   {
      iErr = 1;
      goto Exit;
   }

   /*** set the channel priority to new volume ***/
   /* rm old bits */
   *phEhChnVol &= ~(EH_VOL_MASK << (EH_BITS_PER_CHAN * iChan));

   /* merge */
   *phEhChnVol |= (iVol & EH_VOL_MASK) << (EH_BITS_PER_CHAN * iChan); 

Exit:
   return(iErr);
}


#ifndef ehGetFun

/***
ehGetFun()
Test Function Tracing Bit.
***/

bool
ehGetFun A1(int, iChan)
{
   bool TraceOn = false; /* false = trace off, true = trace on */

   if (phEhChnVol == NO_CHNVOL)
   {
      goto Exit;
   }

   /*** get the channel priority ***/
   TraceOn=!!((*phEhChnVol >> (EH_BITS_PER_CHAN * iChan)) & EH_FUN_BIT);

Exit:
   return(TraceOn);
}

#endif

/***
ehSetFun()
Set channel function tracing bit in shared memory.
***/

bool              /* previous function tracing value */
ehSetFun A2(      /* set function tracing to specified mode */
   int,  iChan,   /* {0..EH_NUMCHAN} which channel to change */
   bool, NewVal) /* {false|true} new bits for channel */
{
   bool OrgVal = false;

   if (phEhChnVol == NO_CHNVOL) /* not initialized yet */
   {
      goto Exit;
   }

   /*** set/reset the function tracing bit ***/
   OrgVal=!!((*phEhChnVol >> (EH_BITS_PER_CHAN * iChan)) & EH_FUN_BIT);
   if (NewVal && !OrgVal)
   {
      *phEhChnVol |= EH_FUN_BIT << (EH_BITS_PER_CHAN * iChan);
   }
   else if (!NewVal && OrgVal)
   {
      *phEhChnVol &= ~(EH_FUN_BIT << (EH_BITS_PER_CHAN * iChan));
   }


Exit:
   return(OrgVal);
}

#ifndef ehGetMem

/***
ehGetMem()
Test Memory Tracing Bit.
***/

int 
ehGetMem A1(int, iChan)
{
   int iMemVal = 0;

   if (phEhChnVol == NO_CHNVOL)
   {
      goto Exit;
   }

   /*** get the memory value ***/
   iMemVal = (int)(*phEhChnVol >> (EH_BITS_PER_CHAN * iChan)) & EH_MEM_MASK;
   iMemVal >>= EH_BIT_OFFSET;

Exit:
   return(iMemVal);
}

#endif

/***
ehSetMem()
Set channel memory tracing bit in shared memory.
***/

int 
ehSetMem A2(int, iChan, 
            int, iMemVal)
             /* {0..EH_NUMCHAN} which channel to change */
             /* {0..3} new setting for memory tracing bit */
{
   int iErr = 0;

   if (phEhChnVol == NO_CHNVOL)
   {
      iErr = 1;
      goto Exit;
   }

   /*** set the new memory value ***/

   /* rm old bits */
   *phEhChnVol &= ~(EH_MEM_MASK << (EH_BITS_PER_CHAN * iChan));

   /* merge */
   iMemVal <<= EH_BIT_OFFSET;
   *phEhChnVol |= (iMemVal & EH_MEM_MASK) << (EH_BITS_PER_CHAN * iChan); 

Exit:
   return(iErr);
}

#ifndef ehGetOut

/***
ehGetOut
Test Output File bit.
***/

int 
ehGetOut A0()
{
   int iOutVal = 0; /* {1=stdout|2=stderr} */

   if (phEhChnVol == NO_CHNVOL)
   {
      goto Exit;
   }

   /*** get the output file value ***/
   iOutVal = (int)(*phEhChnVol >> (EH_BITS_PER_CHAN * EH_NUMCHAN)) & 1;

   if (iOutVal == 0) iOutVal = 2; /* stderr */

Exit:
   return(iOutVal);
}

#endif

/***
ehSetOut()
Set Output File selection bit.
***/

int 
ehSetOut A1(int, iOutVal)
             /* {1=stdout|2=stderr} new setting for output file selection bit */
{
   int iErr = 0;

   if (phEhChnVol == NO_CHNVOL)
   {
      iErr = 1;
      puts("ehSetOut: error");
      goto Exit;
   }

   /*** set the new output file bit ***/
   if (iOutVal == 2) iOutVal = 0;
   if (iOutVal)
   {
      *phEhChnVol |= 1 << (EH_BITS_PER_CHAN * EH_NUMCHAN);
   }
   else
   {
      *phEhChnVol &= ~(1 << (EH_BITS_PER_CHAN * EH_NUMCHAN));
   }

Exit:
   return(iErr);
}

