/***

ehintern.h -- Error Handler, internal definitions

    Copyright (C) 1997  CSEG, Inc., All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                   

    The GNU General Public License is in the file COPYING which resides
    in the top directory of the distribution tree.

***/

#ifndef EHINTERN_H
#define EHINTERN_H 1
static const char*ehintern_h =
       "$Id: ehintern.h,v 1.1.1.1 1997/09/02 12:18:20 ccobb Exp $";

#ifndef   EH_H
#include "eh.h"
#endif

#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE 1
#endif

#ifndef   EHENV_H
#include "ehenv.h"
#endif

/***
 ***
 *** Required include files
 ***
 ***/

EHBEGINEXTERNC

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#ifndef ehEXCEPTIONHANDLING
#include <setjmp.h>
#endif

/*** environment-specific-dependencies ***/

/* variable-argument support */
#if ehSUNOS
#define SUPPORTS_STDARG 0
#include <varargs.h>
#else
#include <stdarg.h>
#define SUPPORTS_STDARG 1
#endif

#ifdef HAVE_SYS_TYPES_H /* && !ehMX && !ehSTMC */
#include <sys/types.h>
#endif

EHENDEXTERNC

/***
 ***
 *** Common definitions
 ***
 ***/

/*** dual-method function argument definitions ***/

/* non-variable argument list */
#if !ehSUNC
#define A0() (void)
#define A1(t1, n1) (t1 n1)
#define A2(t1, n1, t2, n2) (t1 n1, t2 n2)
#define A3(t1, n1, t2, n2, t3, n3) (t1 n1, t2 n2, t3 n3)
#define A4(t1,n1,t2,n2,t3,n3,t4,n4)\
      (t1 n1,t2 n2,t3 n3,t4 n4)
#define A5(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5)
#define A6(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,t6 n6)
#define A7(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,t6 n6,t7 n7)
#define A8(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,t8,n8)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,t6 n6,t7 n7,t8 n8)
#else
#define A0() ()
#define A1(t1, n1) (n1) t1 n1;
#define A2(t1, n1, t2, n2) (n1, n2) t1 n1; t2 n2;
#define A3(t1, n1, t2, n2, t3, n3) (n1, n2, n3) t1 n1; t2 n2; t3 n3;
#define A4(t1,n1,t2,n2,t3,n3,t4,n4)\
      (n1,n2,n3,n4) t1 n1; t2 n2; t3 n3; t4 n4;
#define A5(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5)\
      (n1,n2,n3,n4,n5) t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;
#define A6(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6)\
      (n1,n2,n3,n4,n5,n6) t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;t6 n6;
#define A7(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7)\
      (n1,n2,n3,n4,n5,n6,n7) t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;t6 n6;t7 n7;
#define A8(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,t8,n8)\
      (n1,n2,n3,n4,n5,n6,n7) t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;t6 n6;t7 n7;t8 n8
#endif

/* variable arguments in PROTOTYPE */
#if SUPPORTS_STDARG
#define VA_START(ap,var) va_start(ap,var)
#else
#define VA_START(ap,var) va_start(ap)
#endif

/* variable argument list in FUNCTION DECLARATION */
#if !ehSUNC && SUPPORTS_STDARG
#define A1V(t1, n1, d) (t1 n1, ...)
#define A2V(t1, n1, t2, n2, d) (t1 n1, t2 n2, ...)
#define A3V(t1, n1, t2, n2, t3, n3, d) (t1 n1, t2 n2, t3 n3, ...)
#define A4V(t1,n1,t2,n2,t3,n3,t4,n4,d)\
      (t1 n1,t2 n2,t3 n3, t4 n4, ...)
#define A5V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,d)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,...)
#define A6V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,d)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,t6 n6,...)
#define A7V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,d)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,t6 n6,t7 n7,...)
#define A8V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,t8,n8,d)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,t6 n6,t7 n7,t8 n8,...)
#define A9V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,t8,n8,t9,n9,d)\
      (t1 n1,t2 n2,t3 n3,t4 n4,t5 n5,t6 n6,t7 n7,t8 n8,t9 n9,...)
#else
#define A1V(t1, n1, d) (n1, va_alist) t1 n1; va_dcl
#define A2V(t1, n1, t2, n2, d) (n1, n2, va_alist) t1 n1; t2 n2; va_dcl
#define A3V(t1, n1, t2, n2, t3, n3, d)\
      (n1, n2, n3, va_alist) t1 n1; t2 n2; t3 n3; va_dcl
#define A4V(t1,n1,t2,n2,t3,n3,t4,n4,d)\
      (n1,n2,n3,n4, va_alist) t1 n1;t2 n2;t3 n3;t4 n4; va_dcl
#define A5V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,d)\
      (n1,n2,n3,n4,n5, va_alist) t1 n1;t2 n2;t3 n3;t4 n4;t5 n5; va_dcl
#define A6V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,d)\
      (n1,n2,n3,n4,n5,n6, va_alist)\
      t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;t6 n6; va_dcl
#define A7V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,d)\
      (n1,n2,n3,n4,n5,n6,n7,va_alist)\
      t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;t6 n6;t7 n7; va_dcl
#define A8V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,t8,n8,d)\
      (n1,n2,n3,n4,n5,n6,n7,n8,va_alist)\
      t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;t6 n6;t7 n7;t8 n8;va_dcl
#define A9V(t1,n1,t2,n2,t3,n3,t4,n4,t5,n5,t6,n6,t7,n7,t8,n8,t9,n9,d)\
      (n1,n2,n3,n4,n5,n6,n7,n8,n9,va_alist)\
      t1 n1;t2 n2;t3 n3;t4 n4;t5 n5;t6 n6;t7 n7;t8 n8;t9 n9;va_dcl
#endif

#if ehSUNOS /* prototypes for functions not prototyped in distributed headers */
int    atoi     A((const char *nptr));
void   bcopy    A((const char *src, char *dst, int length));
int    exit     A((int status));
int    fclose   A((FILE *stream));
int    fflush   A((FILE *stream));
int    fprintf  A((FILE *stream, const char *format, ...));
size_t fread    A((void *ptr, size_t size, size_t nmemb, FILE *stream));
size_t fwrite   A((const void*ptr,size_t size,size_t nmemb,FILE*strm));
int    fscanf   A((FILE *stream, const char *format, ...));
int    fputs    A((const char *s, FILE *stream));
int    isatty   A((int fildes));
char*  memcpy   A((void *s1, const void *s2, size_t n));
void * memmove  A((void *s1, const void *s2, size_t n));
char*  memset   A((void *s, int c, size_t n));
void   perror   A((const char *s));
int    printf   A((const char *format, ...));
int    sscanf   A((char *s, const char *format, ...));
time_t time     A((time_t *timeptr));
int    vfprintf A((FILE *stream, const char *format, va_list arg));
#endif

/*** other abreviations ***/

/* generic NULL pointer */
#ifndef NULL
#define NULL  0
#endif

/* ASCII NUL character */
#ifndef NUL
#define NUL ((char)0)
#endif

/* support for repeat-until construct */
#ifndef repeat
#define repeat   do
#endif

#ifndef until
#define until(x) while(!(x))
#endif

/* some abreviations */
#define SZCHR sizeof(char)
#define SZINT sizeof(int)
#define SZLNG sizeof(long)
#define SZARR(a) (sizeof(a)/sizeof(*a)) /* # elements in an array */

/* minumum */
#ifndef MIN
#define MIN(i,j) ((i) < (j) ? i : j)
#endif

/* maximum */
#ifndef MAX
#define MAX(i,j) ((i) < (j) ? j : i)
#endif

/* absolute value */
#ifndef ABS
#define ABS(i)   ((i) < 0 ? -(i) : i)
#endif

/* boolean inclusive check if variable is in range */
#define IN_RANGE(x, lo, hi)  ((lo) <= (x) && (x) <= (hi))

/* free memory and set pointer to NULL */
#ifndef EHFREE
#define EHFREE(p) free(p),p=NULL
#endif

/* boundary alignment requirements on given data type */
static ALIGNTYP _at = 0;   /* alignment type for this compiler */
#define BOUNDOF(typ)\
      ((char *)&((struct{char C;typ T;}*)&_at)->T-(char *)&_at)

/* report offsets of structure members */
#if !defined(offsetof) && !defined(__STDC__) && !defined(__cplusplus)
#define offsetof(stc,m) ((char *)&((stc *)&_at)->m - (char *)&_at)
#endif

/****   data types abbreviations   ****/

#ifndef ulong
#define ulong unsigned long
#endif

#ifndef uint
#define uint unsigned
#endif

#ifndef ushort
#define ushort unsigned short
#endif

#ifndef schar
#define schar signed char
#endif

#ifndef uchar
#define uchar unsigned char
#endif

#ifndef word
#define word unsigned short
#endif

#if !defined(byte) && !(ehSASC && defined(_DOS_H))
#define byte unsigned char
#endif

#ifndef uint
#define uint unsigned
#endif

#define NOREF(v) (v)=(v) /* suppress 'unreferenced-variable' messages */

#define istext(c) ( isascii(c) && !iscntrl(c) )

/* allocate memory for a new structure */
#define EHNEW(stc) (stc *)calloc(sizeof(stc), 1)

/* disable Control-C Checking */
#define EHSTDOUT (Eh.pFilOut)

#define EH_CTRLC_CHK_OFF() (Eh.ucNoCtrlC = 1)
#define EH_CTRLC_CHK_ON()  (Eh.ucNoCtrlC = 1)

/* disable Bounds Checking */
#define EH_BNDS_CHK_OFF() (Eh.ucNoBounds = 1)
#define EH_BNDS_CHK_ON()  (Eh.ucNoBounds = 0)

/*** Function entry/exit tracing support ***/

/*** function pointer types ***/

typedef void (*pfv_pv)      (const void*);    /* ptr to func ret void */
typedef int  (*pfi_pv)      (const void*);    /* ptr to func ret int */
typedef int  (*pfi_pvpv)    (const void*,const void*);

/*** some public function definitions ***/

extern void (*pfvEhIntInit)A0(); /* internal initialization vector */
extern void (*pfvEhExtInit)A0(); /* external initialization vector */

void  ehOsTerm      A((void));


#define EHFORCECTRLC() ehSigInt(SIGINT)


/*** define EhEndCatch macros ***/

#if ehBC
#define EHCALLBACK _USERENTRY
#else
#define EHCALLBACK
#endif

void ehSigInstall A((void));   /* install SIGINT handler */
void ehSigUninstall A((void)); /* uninstall SIGINT handler */
void EHCALLBACK ehSigInt A((int iSig)); /* to match prototype */


#define EHSHIELDON()  (EhSt.ucShield=1)
#define EHSHIELDOFF() (EhSt.ucShield=0)

/*** these macros can only be used in the exit sequence ***/

#define ehFatal()     (EhSt.scError == -1)   /* error (EHTHROW,EHERROR) occurred in CURRENT function */
#define ehAnyError()  (Eh.Er.u.hNo != 0)  /* any error has occurred in this process */
#define ehJmpError()  (iJmpVal!=0)          /* error inherited from longjmp */
#define ehCtrlC()     (EhSt.ucCtrlC!=0)      /* user pressed control C */

#define OFFSET_WIDTH 23  /* # bits in file offset part of error message code */

/*** error processing macros ***/

void ehFunEnt   A((int,const char*,const char*,const int,int,EHSTATE*));
int  ehCatch     A((EHSTATE*pEhSv));
void ehFunRet    A((int,const char*,const char*,const int,int));
void ehFatalJmp  A((void));
void ehExitJmp   A((void));
void ehPostJmp   A((const char*));
int  ehSetVol    A((int iChan, int iFunBit));
int  ehGetVol    A((int iChan));
void ehHexDmp    A((FILE *pFileOut, const char *pcData, int iNum));
void ehLogTxt    A((const char *pcLogTxt));
void ehPgmEnt    A((void));
void ehPgmExt    A((int iChan));
void*ehDynArr    A((void **, size_t*, const size_t));
uint ehRoundUp   A((unsigned int uiValue));
void ehCallStack A((FILE*));


#if 1 /* 1 = macro, 0 = function */
#define ehIndent(pFile)\
   fprintf(pFile,"%*.*s",EhSt.hLev*3,EhSt.hLev*3,""),fflush(pFile)
#else
void ehIndent  A((FILE *pFile));
#endif

#define EHCHAIN                do{EhSt.ucChain=1;EHBREAK;}while(0)
#define EHCHAIN_RETURN(RetVal) do{EhSt.ucChain=1;EHRETURN(RetVal);}while(0)
#define EHCHAIN_LEAVE()        do{EhSt.ucChain=1;EHLEAVE();}while(0)

#define _ASSERT(EX) \
   ehLogTxt("\n");\
   ehIndent(EHSTDERR);\
   fprintf(EHSTDERR,\
         "EHASSERT..(%s) failed: %s <%s %s() %d>\n",\
         #EX, Eh.pcProgName, __FILE__, _EHFUNC_, __LINE__);\
   fflush(EHSTDERR)
   
#define EHASSERT(EX)               if (!(EX))do{_ASSERT(EX);EHTHROW(EH_ASSERT);}while(0)
#define EHASSERT_RETURN(EX,RetVal) if (!(EX))do{_ASSERT(EX);EHTHROW_RETURN(EH_ASSERT,RetVal);}while(0)
#define EHASSERT_LEAVE(EX)         if (!(EX))do{_ASSERT(EX);EHTHROW_LEAVE(EH_ASSERT);}while(0)

#define EH_SZBUF 81


#ifndef ehEXCEPTIONHANDLING /*** C model exception handling ***/


/* _FUN_TRACE() is used by both FUN_TRACE() and PGM_TRACE() */

#ifdef EH_SUPPLY_FUNCTION_NAME
#define _FUN_TRACE(pcFunc)\
   EHSIMCTRLC();\
   ehFunEnt(__CHAN__,__FILE__,pcFunc,__LINE__,CHKOVR,&EhSv)
#else  /* don't need to EH_SUPPLY_FUNCTION_NAME */
#define _FUN_TRACE()\
   EHSIMCTRLC();\
   ehFunEnt(__CHAN__,__FILE__,__FUNC__,__LINE__,CHKOVR,&EhSv)
#endif /* don't need to EH_SUPPLY_FUNCTION_NAME */

/* FUN_TRACE() is the function tracing entry point */

#ifdef EH_SUPPLY_FUNCTION_NAME
#define FUN_TRACE(pcFunc)\
      const char *volatile _EHFUNC_ = pcFunc;\
      EHSTATE EhSv;\
      _FUN_TRACE(_EHFUNC_)
#else  /* don't need to EH_SUPPLY_FUNCTION_NAME */
#define FUN_TRACE()\
      const char *volatile _EHFUNC_ = __FUNC__;\
      EHSTATE EhSv;\
      _FUN_TRACE(_EHFUNC_)
#endif /* don't need to EH_SUPPLY_FUNCTION_NAME */

/* EHLEAVE() is required for functions returning void */

#define EHLEAVE()\
 do{ehFunRet(__CHAN__,__FILE__,_EHFUNC_,__LINE__,CHKOVR);return;}while(0)

/* EHRETURN() is required for functions returning a value */

#define EHRETURN(RetVal)\
   return (ehFunRet(__CHAN__,__FILE__,_EHFUNC_,__LINE__,CHKOVR),RetVal)

#define _FUN_CATCH()\
 if (ehCatch(&EhSv)) EHBREAK;\
 if (setjmp(EhSt.JmpBuf))\
 {\
  ehPostJmp(_EHFUNC_);\
 }\
 else\
 {

#ifdef EH_SUPPLY_FUNCTION_NAME
#define FUN_CATCH(pcFunc)\
 FUN_TRACE(pcFunc);\
 _FUN_CATCH()
#else  /* don't need to EH_SUPPLY_FUNCTION_NAME */
#define FUN_CATCH()\
 FUN_TRACE();\
 _FUN_CATCH()
#endif /* don't need to EH_SUPPLY_FUNCTION_NAME */

#define ExitSequence \
}_ExitSequence:\
EHINHIBIT();\
__ExitSequence

#define EhCatchPoint \
}_ExitSequence:\
EHINHIBIT();\
__ExitSequence

/* _PGM_TRACE() is the core definition for PGM_TRACE() */

#define _PGM_TRACE() \
   EHSTATE EhSv;\
   Eh.pcProgName = (char *)argv[0];argc=argc;\
   ehPgmEnt()

/* PGM_TRACE() performs function tracing initilization in main */

#ifndef _EHFUNC_
/* we use an explicit "main" here to later check if we are in main() */
#define PGM_TRACE()\
 const char *_EHFUNC_="main";\
 _PGM_TRACE();\
 _FUN_CATCH();\
 _FUN_TRACE(_EHFUNC_)
#else
#define PGM_TRACE()\
 _PGM_TRACE();\
 _FUN_CATCH();\
 _FUN_TRACE(_EHFUNC_)
#endif

/* PGM_RETURN() is required: you should *always* return a value to the OS */

#define PGM_RETURN(RetVal)\
   return(\
      ehOsTerm(),\
      ehFunRet(__CHAN__,__FILE__,_EHFUNC_,__LINE__,CHKOVR),\
      ehPgmExt(__CHAN__),\
      RetVal)


#else /*** ehEXCEPTIONHANDLING *** C++ model exception handling ***/

class EHMEFIRST
{
public:
   EHMEFIRST();   // wrapper
   ~EHMEFIRST();  // object
};

// eXception Handler support

extern char cEhFunTrc; // internal function tracing
extern char cEhPgmTrc; // internal program  tracing

// replace the eh version of function tracing macros with xh version


#endif /* ehEXCEPTIONHANDLING */


/*** former es.h -- Error-handling Support library header ***/


/*** support for ehlist.c (linked list) ***/


bool    ehAddHead     A((EHLIST *pEhList, EHNODE *pEhNode));
bool    ehAddTail     A((EHLIST *pEhList, EHNODE *pEhNode));
void    ehInsert      A((EHLIST *pEhList, EHNODE *pEhNodeNew, EHNODE *pEhNodePred));
bool    ehListIsEmpty A((EHLIST *pEhList));
void    ehNewList     A((EHLIST *pEhList));
EHNODE *ehRemHead     A((EHLIST *pEhList));
EHNODE *ehInqHead     A((EHLIST *pEhList));
bool    ehRemove      A((EHNODE *pEhNode));
EHNODE *ehRemTail     A((EHLIST *pEhList));
EHNODE *ehLocTail     A((EHLIST *pEhList));
#define ehLocPred(pEhNode) ((pEhNode)->pEhNodePred)

/* linked list (named, prioritized) support */
void    ehEnqueue     A((EHLIST *pEhList, EHNODE *pEhNode));
EHNODE *ehFindName    A((EHLIST *pEhList, byte *pbName));

/* traversing a list */
#define ehListHead(pEhList) ((pEhList)->pEhNodeHead)
#define ehListEnd(pEhNode)  ((pEhNode)->pEhNodeSucc == 0)
#define ehListNext(pEhNode) ((pEhNode)->pEhNodeSucc)


/*** support for ehavltre.c (avl tree) ***/

extern char cEhAvlDmp;

typedef void (*pfv_ppEhAvlNode)(EHAVLNODE**);

EHAVL*ehAvlInit     A((pfi_pvpv pfiCompare, pfi_ppv pfiDelete,
                        pfv_pFilepv pfvDump));
void *ehAvlSrch     A((EHAVL *pEhAvl, const void *pvUserData));
bool  ehAvlAdd      A((EHAVL *pEhAvl, const void *pvUserData,
                        pfi_pvpv pfiCompare));
int   ehAvlDel      A((EHAVL *pEhAvl, const void *pvUserData,
                        pfi_pvpv pfiCompare));
void  ehAvlTerm     A((EHAVL *pEhAvl));
int   ehAvlTrav     A((EHAVL*pEhAvl, pfv_ppEhAvlNode pfvAvlNodeFunc,
                        pfi_ppv pfiUserFunc));
void  ehAvlDump     A((EHAVL*pEhAvl, FILE *pFile));
void  ehAvlTreeDump A((FILE*, EHAVL*pEhAvl));

/*** support for ehchkres.c (check resource tracking results) ***/

extern char cEhResDmp;
extern char cEhFunDmp;

void    ehResInit     A((EHRESTRK *pEhResTrk));
void    ehResTerm     A((EHRESTRK *pEhResTrk));
void    ehResDmpAvl   A((EHAVL* pEhAvl, int iLevel));

/*** support for ehtrkres.c (resource tracking) ***/

typedef ulong (*pful_ulv)(ulong,...); /* ptr to func ret ulong, var args */

#define EHRESLIST struct _eh_reslist

/* keep track of multiple CALLs to obtain same RESource */
typedef struct
{
   EHNODE      EhNode;     /* linked list node */
   EHRESLIST  *pEhResList; /* pointer back to parent */
   int         iObtFun;    /* actual resource-obtaining func */
   ulong      *pulArgs;    /* args used when obtaining this resource */
   const char *pcFileO;    /* source file where resource was obtained */
   const char *pcFuncO;    /* function in which resource was obtained */
   int         iLineO;     /* line on which resource was obtained */
   ulong       ulCountO;   /* which time through this statement */
   const char *pcFileR;    /* source file where resource was released */
   const char *pcFuncR;    /* function in which resource was released */
   int         iLineR;     /* line on which resource was released */
   ulong       ulCountR;   /* which time through this statement */
} EHRESCALL, RC;           /* track resource function calls */

typedef enum 
{
   EH_RES_NONE,
   EH_RES_RELEASE,
   EH_RES_REOBTAIN,
   EH_RES_OBTAIN
} EHRESWHOM;        /* who most recently operated on the resource */

typedef enum 
{
   EH_RES_UNUSED,
   EH_RES_RELEASED, /* have released the resource */
   EH_RES_OBTAINED  /* have obtained the resource */
} EHRESSTATE;       /* state of resource */

/* contains attributes of each resource in use */
EHRESLIST
{
   int        iType;      /* unique resource type */
   char      *pcResNam;   /* name of resource */
   pfv_ulv    pfvDefRel;  /* ptr to default resource-releasing func */
   pfv_ulv    pfvActRel;  /* ptr to actual  resource-releasing func */
   ulong      ulFlags;    /* define some resource characteristics */
   ulong      ulResHndl;  /* resource handle */
   EHRESWHOM  EhResWhom;  /* who most recently operated on resource */
   EHRESSTATE EhResState; /* state of resource */
   bool       hasProblem; /* problem discovered with resource */
   ushort     uhNumObtn;  /* # RESCALL structs OBTAINED by this res */
   size_t     szRCInUse;  /* # RESCALL structures IN USE */
   size_t     szRCAlloc;  /* # RESCALL structures allocated */
   EHRESCALL *pEhResCall; /* keep track of function calls */
   EHRESLIST *pEhRLOnRel; /* related resource to release 
                             when this one is released */
};

/* line and file where resource obtaining function was called */
typedef struct _eh_res_func_
{
   int   iLine;        /* line where obtaining function was called */
   const char *pcFile; /* file where obtaining function was called */
   ulong ulCount;      /* # times this function has been called */
} EHRESFUNC;

#if 0
/* release(EHRESLIST *) for customization */
#define EH_RES_REL_RESLIST (1L << (EH_RES_NUMARGS_WID + 5))
/* call: release(hndl, NULL) */
#define EH_RES_REL_H_NULL  (1L << (EH_RES_NUMARGS_WID + 6))
/* call: release(hndl, arg1) */
#define EH_RES_REL_H_ARG1  (1L << (EH_RES_NUMARGS_WID + 7))
#endif

pfv_ulv ehResRelBy    A((EHRESTRK *pEhResTrk, ulong ulHndl, int iType));
void    ehResSetRelBy A((EHRESTRK *pEhResTrk, ulong ulHndl, int iType,
                        pfv_ulv pfvRelBy));

#endif /* EHINTERN_H */
