/***

ehremote.c -- Error Handler, remote control of EH applications

    Copyright (C) 1997  CSEG, Inc., All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                   

    The GNU General Public License is in the file COPYING which resides
    in the top directory of the distribution tree.

***/

static const char *rcsid =
       "$Id: ehremote.c,v 1.1.1.1 1997/09/02 12:18:21 ccobb Exp $";

#include <string.h>
#include <ctype.h>

#include "eh.h"
#include "eherrs.h"
#include "ehchnvol.h"
#include "ehvolume.h"

EHFILENAME("ehremote.c");

static
void
print_stats()
{
   int iChan;

   printf("\t-o%d ", ehGetOut());
   for (iChan = 0; iChan < EH_NUMCHAN; ++iChan)
   {
      printf("-c%d ", iChan);
      printf("-v%c ", tolower(*apcEhVol[ehGetVol(iChan)]));
      printf("-f%d ", ehGetFun(iChan));
      //printf("-m%d",   ehGetMem(iChan));
   }
   puts("");
}

int 
main(
   int    argc, 
   char** argv)
{
   short *volatile phRemote = NO_CHNVOL;
   volatile int iShmId = 0;
   short hInitialSettings;

   if (argc == 1)
   {
      puts("usage: ehremote -p <process id> { -o[0|1] -c[0..2] -v[0..2] -f[0|1] }...");
      exit(1);
   }

   int iPID = -1;
   int iChan; /* channel */
   int iVol;  /* volume on a channel */
   bool funTrc;  /* function tracing state on channel */
   int iMem;  /* memory tracing debug level state on channel */
   int iOut;  /* global stderr output stream: stderr or log file */

   // process the command line
   NOREF(argc); // no reference to argc
   while (*++argv != NULL)
   {
      char *pcSetting = *argv;

      // check for process id
      if (strncmp(pcSetting, "-p", 2) == 0)
      {
         pcSetting += 2;
         if (*pcSetting == NUL)
         {
            pcSetting = *++argv;
         }
         iPID = atoi(pcSetting);
         if (ehInitChnVol(iPID)) // phChnVol now points to shared memory
         {
            fprintf(stderr, "ehInitChnVol(%d) failed\n", iPID);
            break;
         }
         hInitialSettings = *phEhChnVol;
         printf("Initial status of process %d:", iPID);
         print_stats();
         iChan =  0; /* by default, operate on channel 0 */
         continue;
      }

      if (ehSetVal(pcSetting, &iChan, &iVol, &funTrc, &iMem, &iOut))
      {
         fprintf(EHSTDERR, "bad EHVOLUME option: %s\n", pcSetting);
         fflush(EHSTDERR);
         continue;
      }
   } // end while

   if (phEhChnVol != NO_CHNVOL && *phEhChnVol != hInitialSettings)
   {
      printf("Final status  of  process %d:", iPID);
      print_stats();
   }
   ehTermChnVol();
   return 0;
}
